/* GTK - The GIMP Toolkit
 * Copyright (C) 2000 Red Hat Software
 * Copyright (C) 2002 Yusuke Tabata
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Owen Taylor <otaylor@redhat.com>
 *          Yusuke Tabata <tee@kuis.kyoto-u.ac.jp>
 *          Yukihiro Nakai <ynakai@redhat.com>
 *
 */

#include <string.h>

#include <gdk/gdkkeysyms.h>

#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>

#define IM_CONTEXT_NAKAI(obj) (GTK_CHECK_CAST((obj), type_nakai, IMContextNakai))
#define BUFSIZ 1024

typedef struct _IMContextNakai {
  struct _GtkIMContext parent;
  guchar* preedit_buf; /* Hiragana string as much as possible */
} IMContextNakai;

typedef struct _IMContextNakaiClass {
  GtkIMContextClass parent_class;
} IMContextNakaiClass;

typedef struct _romapair {
  gchar* roma;
  gchar* kana;
} romapair;

#include "convtable.h"

GType type_nakai = 0;

static void im_nakai_class_init (GtkIMContextClass *class);
static void im_nakai_init (GtkIMContext *im_context);
static gboolean im_nakai_filter_keypress(GtkIMContext *context,
		                GdkEventKey *key);
static void im_nakai_get_preedit_string(GtkIMContext *, gchar **str,
					PangoAttrList **attrs,
					gint *cursor_pos);
static guchar* euc2utf8(const guchar* str);

int
g_strrncmp(gchar* s1, gchar* s2, size_t n) {
  gchar* s1tmp = s1;
  if( strlen(s1) >= n ) {
    s1tmp += strlen(s1)-n;
  }

  return strcmp(s1tmp, s2);
}

gchar*
roma2kana(gchar* romastr) {
  gchar* result = NULL;
  int tsize = sizeof(kanatable)/sizeof(kanatable[0]);
  int i;

  if( romastr == NULL ) return NULL;
  if( *romastr == '\0' ) return NULL;

  for(i=0;i<tsize;i++) {
    romapair rp = kanatable[i];
    if( g_strrncmp(romastr, rp.roma, strlen(rp.roma)) == 0 ) {
      int len = strlen(romastr)-strlen(rp.roma)+strlen(rp.kana);
      result = g_new0(gchar, len+1);
      strncat(result, romastr, strlen(romastr)-strlen(rp.roma));
      strcat(result, rp.kana);
      return result;
    }
  }

  result = g_strdup(romastr); /* No change */

  return result;
}

void
buffer_append(gchar* buf, gchar ch) {
  gchar* kana = NULL;

  if( buf == NULL ) return;

  buf[strlen(buf)] = ch;
  buf[strlen(buf)] = '\0';
  
  kana = roma2kana(buf);
  strcpy(buf, kana);
  g_free(kana);
}

static void
im_nakai_register_type (GTypeModule *module)
{
  static const GTypeInfo object_info =
  {
    sizeof (IMContextNakaiClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) im_nakai_class_init,
    NULL,           /* class_finalize */
    NULL,           /* class_data */
    sizeof (IMContextNakai),
    0,
    (GtkObjectInitFunc) im_nakai_init,
  };

  type_nakai = 
    g_type_module_register_type (module,
				 GTK_TYPE_IM_CONTEXT,
				 "IMContextNakai",
				 &object_info, 0);
}

static void
im_nakai_class_init (GtkIMContextClass *class)
{
  GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(class);
  im_context_class->filter_keypress = im_nakai_filter_keypress;
  im_context_class->get_preedit_string = im_nakai_get_preedit_string;
}

static void
im_nakai_init (GtkIMContext *im_context)
{
  IMContextNakai *cn = IM_CONTEXT_NAKAI(im_context);
  cn->preedit_buf = g_new0(guchar, BUFSIZ);

}

static const GtkIMContextInfo im_nakai_info = { 
  "Nakai",		   /* ID */
  "Hiragana",             /* Human readable name */
  "gtk+",			   /* Translation domain */
   "",		   /* Dir for bindtextdomain (not strictly needed for "gtk+") */
   ""			           /* Languages for which this module is the default */
};

static const GtkIMContextInfo *info_list[] = {
  &im_nakai_info
};

void
im_module_init (GTypeModule *module)
{
  im_nakai_register_type(module);
}

void 
im_module_exit (void)
{
}

void 
im_module_list (const GtkIMContextInfo ***contexts,
		int                      *n_contexts)
{
  *contexts = info_list;
  *n_contexts = G_N_ELEMENTS (info_list);
}

GtkIMContext *
im_module_create (const gchar *context_id)
{
  if (strcmp (context_id, "Nakai") == 0)
    return GTK_IM_CONTEXT (g_object_new (type_nakai, NULL));
  else
    return NULL;
}

gboolean
im_nakai_filter_keypress(GtkIMContext *context, GdkEventKey *key)
{
  IMContextNakai *cn = IM_CONTEXT_NAKAI(context);

  if( key->type == GDK_KEY_RELEASE ) {
    return FALSE;
  };
  if( key->keyval == GDK_Return ) {
    g_signal_emit_by_name(cn, "commit", euc2utf8(cn->preedit_buf));
    memset(cn->preedit_buf, 0, BUFSIZ);
    g_signal_emit_by_name(cn, "preedit_changed");
  } else if( key->keyval >= GDK_a && key->keyval <= GDK_z ) {
    buffer_append(cn->preedit_buf, gdk_keyval_name(key->keyval)[0]);
    g_signal_emit_by_name(cn, "preedit_changed");
  }

  return TRUE;
}

void
im_nakai_get_preedit_string(GtkIMContext *ic, gchar **str,
			    PangoAttrList **attrs, gint *cursor_pos)
{
  IMContextNakai *cn = IM_CONTEXT_NAKAI(ic);
  if (attrs) {
    *attrs = pango_attr_list_new();
  }
  if (cursor_pos) {
    *cursor_pos = 0;
  }
  *str = euc2utf8(cn->preedit_buf);

  if( *str == NULL || strlen(*str) == 0 ) return;

  if (attrs) {
    PangoAttribute *attr;
    attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
    attr->start_index = 0;
    attr->end_index = strlen(*str);
    pango_attr_list_insert(*attrs, attr);
  }

  *cursor_pos = strlen(*str);
}

static guchar *
euc2utf8 (const guchar *str)
{
  GError *error = NULL;
  gchar *result;

  result = g_convert (str, -1,
		      "UTF-8", "EUC-JP",
		      NULL, NULL, &error);
  if (!result)
    {
      g_warning ("Error converting text from IM to UTF-8: %s\n", error->message);
      g_error_free (error);
    }

  return result;
}
